@@ -1,9 +1,9 @@
 /*
   ziptool.c -- tool for modifying zip archive in multiple ways
-  Copyright (C) 2012-2020 Dieter Baron and Thomas Klausner
+  Copyright (C) 2012-2022 Dieter Baron and Thomas Klausner
 
   This file is part of libzip, a library to manipulate ZIP archives.
-  The authors can be contacted at <libzip@nih.at>
+  The authors can be contacted at <info@libzip.org>
 
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
@@ -33,6 +33,8 @@
 
 #include "config.h"
 
+#include "compat.h"
+
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
@@ -55,21 +57,22 @@ extern int optopt;
 
 #include "zip.h"
 
-#include "compat.h"
-
 typedef struct dispatch_table_s {
     const char *cmdline_name;
     int argument_count;
     const char *arg_names;
     const char *description;
-    int (*function)(int argc, char *argv[]);
+    int (*function)(char *argv[]);
 } dispatch_table_t;
 
 static zip_flags_t get_flags(const char *arg);
 static zip_int32_t get_compression_method(const char *arg);
 static zip_uint16_t get_encryption_method(const char *arg);
 static void hexdump(const zip_uint8_t *data, zip_uint16_t len);
+static int parse_archive_flag(const char* arg);
 int ziptool_post_close(const char *archive);
+static const char* decode_filename(const char* name);
+static const char* encode_filename(const char* name);
 
 #ifndef FOR_REGRESS
 #define OPTIONS_REGRESS ""
@@ -79,9 +82,72 @@ int ziptool_post_close(const char *archive);
 zip_t *za, *z_in[16];
 unsigned int z_in_count;
 zip_flags_t stat_flags;
+int hex_encoded_filenames = 0; // Can only be set in ziptool_regress.
 
 static int
-add(int argc, char *argv[]) {
+cat_impl(zip_uint64_t idx, zip_uint64_t start, zip_uint64_t len) {
+    zip_error_t error;
+    zip_source_t *src;
+    zip_int64_t n;
+    char buf[8192];
+
+#ifdef _WIN32
+    /* Need to set stdout to binary mode for Windows */
+    setmode(fileno(stdout), _O_BINARY);
+#endif
+    zip_error_init(&error);
+    /* we can't pass 0 as a len to zip_source_zip_create because it
+       will try to give us compressed data */
+    if (len == 0) {
+        struct zip_stat sb;
+
+        if (zip_stat_index(za, idx, stat_flags, &sb) < 0) {
+            fprintf(stderr, "zip_stat_index failed on '%" PRIu64 "' failed: %s\n", idx, zip_strerror(za));
+            return -1;
+        }
+
+        if (!(sb.valid & ZIP_STAT_SIZE)) {
+            fprintf(stderr, "can't cat file at index '%" PRIu64 "' with unknown size\n", idx);
+            return -1;
+        }
+        len = sb.size;
+    }
+    if ((src = zip_source_zip_file_create(za, idx, 0, start, len, NULL, &error)) == NULL) {
+        fprintf(stderr, "can't open file at index '%" PRIu64 "': %s\n", idx, zip_error_strerror(&error));
+        zip_error_fini(&error);
+        return -1;
+    }
+    zip_error_fini(&error);
+
+    if (zip_source_open(src) < 0) {
+        fprintf(stderr, "can't open file at index '%" PRIu64 "': %s\n", idx, zip_error_strerror(zip_source_error(src)));
+        zip_source_free(src);
+        return -1;
+    }
+    while ((n = zip_source_read(src, buf, sizeof(buf))) > 0) {
+        if (fwrite(buf, (size_t)n, 1, stdout) != 1) {
+            fprintf(stderr, "can't write file contents to stdout: %s\n", strerror(errno));
+            zip_source_free(src);
+            return -1;
+        }
+    }
+    if (n == -1) {
+        fprintf(stderr, "can't read file at index '%" PRIu64 "': %s\n", idx, zip_error_strerror(zip_source_error(src)));
+        zip_source_free(src);
+        return -1;
+    }
+    if (zip_source_close(src) < 0) {
+        fprintf(stderr, "can't close file at index '%" PRIu64 "': %s\n", idx, zip_error_strerror(zip_source_error(src)));
+        zip_source_free(src);
+        return -1;
+    }
+    zip_source_free(src);
+
+    return 0;
+}
+
+static int
+add(char *argv[]) {
     zip_source_t *zs;
 
     if ((zs = zip_source_buffer(za, argv[1], strlen(argv[1]), 0)) == NULL) {
@@ -89,7 +155,7 @@ add(int argc, char *argv[]) {
         return -1;
     }
 
-    if (zip_add(za, argv[0], zs) == -1) {
+    if (zip_file_add(za, decode_filename(argv[0]), zs, 0) == -1) {
         zip_source_free(zs);
         fprintf(stderr, "can't add file '%s': %s\n", argv[0], zip_strerror(za));
         return -1;
@@ -98,9 +164,9 @@ add(int argc, char *argv[]) {
 }
 
 static int
-add_dir(int argc, char *argv[]) {
+add_dir(char *argv[]) {
     /* add directory */
-    if (zip_add_dir(za, argv[0]) < 0) {
+    if (zip_dir_add(za, decode_filename(argv[0]), 0) < 0) {
         fprintf(stderr, "can't add directory '%s': %s\n", argv[0], zip_strerror(za));
         return -1;
     }
@@ -108,7 +174,7 @@ add_dir(int argc, char *argv[]) {
 }
 
 static int
-add_file(int argc, char *argv[]) {
+add_file(char *argv[]) {
     zip_source_t *zs;
     zip_uint64_t start = strtoull(argv[2], NULL, 10);
     zip_int64_t len = strtoll(argv[3], NULL, 10);
@@ -126,7 +192,7 @@ add_file(int argc, char *argv[]) {
         }
     }
 
-    if (zip_add(za, argv[0], zs) == -1) {
+    if (zip_file_add(za, decode_filename(argv[0]), zs, 0) == -1) {
         zip_source_free(zs);
         fprintf(stderr, "can't add file '%s': %s\n", argv[0], zip_strerror(za));
         return -1;
@@ -135,11 +201,12 @@ add_file(int argc, char *argv[]) {
 }
 
 static int
-add_from_zip(int argc, char *argv[]) {
+add_from_zip(char *argv[]) {
     zip_uint64_t idx, start;
     zip_int64_t len;
     int err;
     zip_source_t *zs;
+    zip_flags_t flags = 0;
     /* add from another zip file */
     idx = strtoull(argv[2], NULL, 10);
     start = strtoull(argv[3], NULL, 10);
@@ -151,12 +218,15 @@ add_from_zip(int argc, char *argv[]) {
         zip_error_fini(&error);
         return -1;
     }
-    if ((zs = zip_source_zip(za, z_in[z_in_count], idx, 0, start, len)) == NULL) {
+    if (start == 0 && len == -1) {
+        flags = ZIP_FL_COMPRESSED;
+    }
+    if ((zs = zip_source_zip_file(za, z_in[z_in_count], idx, flags, start, len, NULL)) == NULL) {
         fprintf(stderr, "error creating file source from '%s' index '%" PRIu64 "': %s\n", argv[1], idx, zip_strerror(za));
         zip_close(z_in[z_in_count]);
         return -1;
     }
-    if (zip_add(za, argv[0], zs) == -1) {
+    if (zip_file_add(za, decode_filename(argv[0]), zs, 0) == -1) {
         fprintf(stderr, "can't add file '%s': %s\n", argv[0], zip_strerror(za));
         zip_source_free(zs);
         zip_close(z_in[z_in_count]);
@@ -167,48 +237,30 @@ add_from_zip(int argc, char *argv[]) {
 }
 
 static int
-cat(int argc, char *argv[]) {
+cat(char *argv[]) {
     /* output file contents to stdout */
     zip_uint64_t idx;
-    zip_int64_t n;
-    zip_file_t *zf;
-    char buf[8192];
-    int err;
     idx = strtoull(argv[0], NULL, 10);
 
-#ifdef _WIN32
-    /* Need to set stdout to binary mode for Windows */
-    setmode(fileno(stdout), _O_BINARY);
-#endif
-    if ((zf = zip_fopen_index(za, idx, 0)) == NULL) {
-        fprintf(stderr, "can't open file at index '%" PRIu64 "': %s\n", idx, zip_strerror(za));
-        return -1;
-    }
-    while ((n = zip_fread(zf, buf, sizeof(buf))) > 0) {
-        if (fwrite(buf, (size_t)n, 1, stdout) != 1) {
-            zip_fclose(zf);
-            fprintf(stderr, "can't write file contents to stdout: %s\n", strerror(errno));
-            return -1;
-        }
-    }
-    if (n == -1) {
-        fprintf(stderr, "can't read file at index '%" PRIu64 "': %s\n", idx, zip_file_strerror(zf));
-        zip_fclose(zf);
-        return -1;
-    }
-    if ((err = zip_fclose(zf)) != 0) {
-        zip_error_t error;
-
-        zip_error_init_with_code(&error, err);
-        fprintf(stderr, "can't close file at index '%" PRIu64 "': %s\n", idx, zip_error_strerror(&error));
-        return -1;
-    }
-
-    return 0;
+    return cat_impl(idx, 0, 0);
 }
 
 static int
-count_extra(int argc, char *argv[]) {
+cat_partial(char *argv[]) {
+    /* output partial file contents to stdout */
+    zip_uint64_t idx;
+    zip_uint64_t start;
+    zip_uint64_t len;
+
+    idx = strtoull(argv[0], NULL, 10);
+    start = strtoull(argv[1], NULL, 10);
+    len = strtoull(argv[2], NULL, 10);
+
+    return cat_impl(idx, start, len);
+}
+
+static int
+count_extra(char *argv[]) {
     zip_int16_t count;
     zip_uint64_t idx;
     zip_flags_t ceflags = 0;
@@ -225,7 +277,7 @@ count_extra(int argc, char *argv[]) {
 }
 
 static int
-count_extra_by_id(int argc, char *argv[]) {
+count_extra_by_id(char *argv[]) {
     zip_int16_t count;
     zip_uint16_t eid;
     zip_flags_t ceflags = 0;
@@ -243,7 +295,7 @@ count_extra_by_id(int argc, char *argv[]) {
     return 0;
 }
 
-static int delete (int argc, char *argv[]) {
+static int delete (char *argv[]) {
     zip_uint64_t idx;
     idx = strtoull(argv[0], NULL, 10);
     if (zip_delete(za, idx) < 0) {
@@ -254,7 +306,7 @@ static int delete (int argc, char *argv[]) {
 }
 
 static int
-delete_extra(int argc, char *argv[]) {
+delete_extra(char *argv[]) {
     zip_flags_t geflags;
     zip_uint16_t eid;
     zip_uint64_t idx;
@@ -269,7 +321,7 @@ delete_extra(int argc, char *argv[]) {
 }
 
 static int
-delete_extra_by_id(int argc, char *argv[]) {
+delete_extra_by_id(char *argv[]) {
     zip_flags_t geflags;
     zip_uint16_t eid, eidx;
     zip_uint64_t idx;
@@ -285,19 +337,31 @@ delete_extra_by_id(int argc, char *argv[]) {
 }
 
 static int
-get_archive_comment(int argc, char *argv[]) {
+get_archive_comment(char *argv[]) {
     const char *comment;
     int len;
     /* get archive comment */
-    if ((comment = zip_get_archive_comment(za, &len, 0)) == NULL)
+    if ((comment = zip_get_archive_comment(za, &len, 0)) == NULL || len == 0)
         printf("No archive comment\n");
     else
-        printf("Archive comment: %.*s\n", len, comment);
+        printf("Archive comment: %.*s\n", len, encode_filename(comment));
     return 0;
 }
 
 static int
-get_extra(int argc, char *argv[]) {
+get_archive_flag(char *argv[]) {
+    int flag = parse_archive_flag(argv[0]);
+    if (flag < 0) {
+        fprintf(stderr, "invalid archive flag '%s'\n", argv[0]);
+        return -1;
+    }
+
+    printf("%d\n", zip_get_archive_flag(za, flag, 0));
+    return 0;
+}
+
+static int
+get_extra(char *argv[]) {
     zip_flags_t geflags;
     zip_uint16_t id, eidx, eflen;
     const zip_uint8_t *efdata;
@@ -320,7 +384,7 @@ get_extra(int argc, char *argv[]) {
 }
 
 static int
-get_extra_by_id(int argc, char *argv[]) {
+get_extra_by_id(char *argv[]) {
     zip_flags_t geflags;
     zip_uint16_t eid, eidx, eflen;
     const zip_uint8_t *efdata;
@@ -343,25 +407,25 @@ get_extra_by_id(int argc, char *argv[]) {
 }
 
 static int
-get_file_comment(int argc, char *argv[]) {
+get_file_comment(char *argv[]) {
     const char *comment;
-    int len;
+    zip_uint32_t len;
     zip_uint64_t idx;
     /* get file comment */
     idx = strtoull(argv[0], NULL, 10);
-    if ((comment = zip_get_file_comment(za, idx, &len, 0)) == NULL) {
+    if ((comment = zip_file_get_comment(za, idx, &len, 0)) == NULL) {
         fprintf(stderr, "can't get comment for '%s': %s\n", zip_get_name(za, idx, 0), zip_strerror(za));
         return -1;
     }
     else if (len == 0)
         printf("No comment for '%s'\n", zip_get_name(za, idx, 0));
     else
-        printf("File comment for '%s': %.*s\n", zip_get_name(za, idx, 0), len, comment);
+        printf("File comment for '%s': %.*s\n", zip_get_name(za, idx, 0), (int)len, comment);
     return 0;
 }
 
 static int
-get_num_entries(int argc, char *argv[]) {
+get_num_entries(char *argv[]) {
     zip_int64_t count;
     zip_flags_t flags;
     /* get number of entries in archive */
@@ -372,12 +436,12 @@ get_num_entries(int argc, char *argv[]) {
 }
 
 static int
-name_locate(int argc, char *argv[]) {
+name_locate(char *argv[]) {
     zip_flags_t flags;
     zip_int64_t idx;
     flags = get_flags(argv[1]);
 
-    if ((idx = zip_name_locate(za, argv[0], flags)) < 0) {
+    if ((idx = zip_name_locate(za, decode_filename(argv[0]), flags)) < 0) {
         fprintf(stderr, "can't find entry with name '%s' using flags '%s'\n", argv[0], argv[1]);
     }
     else {
@@ -401,16 +465,16 @@ progress_callback(zip_t *archive, double percentage, void *ud) {
 }
 
 static int
-print_progress(int argc, char *argv[]) {
+print_progress(char *argv[]) {
     zip_register_progress_callback_with_state(za, 0.001, progress_callback, NULL, NULL);
     return 0;
 }
 
 static int
-zrename(int argc, char *argv[]) {
+zrename(char *argv[]) {
     zip_uint64_t idx;
     idx = strtoull(argv[0], NULL, 10);
-    if (zip_rename(za, idx, argv[1]) < 0) {
+    if (zip_file_rename(za, idx, decode_filename(argv[1]), 0) < 0) {
         fprintf(stderr, "can't rename file at index '%" PRIu64 "' to '%s': %s\n", idx, argv[1], zip_strerror(za));
         return -1;
     }
@@ -418,7 +482,7 @@ zrename(int argc, char *argv[]) {
 }
 
 static int
-replace_file_contents(int argc, char *argv[]) {
+replace_file_contents(char *argv[]) {
     /* replace file contents with data from command line */
     const char *content;
     zip_source_t *s;
@@ -434,7 +498,7 @@ replace_file_contents(int argc, char *argv[]) {
 }
 
 static int
-set_extra(int argc, char *argv[]) {
+set_extra(char *argv[]) {
     zip_flags_t geflags;
     zip_uint16_t eid, eidx;
     const zip_uint8_t *efdata;
@@ -452,7 +516,7 @@ set_extra(int argc, char *argv[]) {
 }
 
 static int
-set_archive_comment(int argc, char *argv[]) {
+set_archive_comment(char *argv[]) {
     if (zip_set_archive_comment(za, argv[0], (zip_uint16_t)strlen(argv[0])) < 0) {
         fprintf(stderr, "can't set archive comment to '%s': %s\n", argv[0], zip_strerror(za));
         return -1;
@@ -461,7 +525,25 @@ set_archive_comment(int argc, char *argv[]) {
 }
 
 static int
-set_file_comment(int argc, char *argv[]) {
+set_archive_flag(char *argv[]) {
+    int flag = parse_archive_flag(argv[0]);
+    if (flag < 0) {
+        fprintf(stderr, "invalid archive flag '%s'\n", argv[0]);
+        return -1;
+    }
+
+    int value = strcasecmp(argv[1], "1") == 0 || strcasecmp(argv[1], "true") == 0 || strcasecmp(argv[1], "yes") == 0;
+
+    if (zip_set_archive_flag(za, flag, value) < 0) {
+        fprintf(stderr, "can't set archive flag '%s' to %d: %s\n", argv[0], value, zip_strerror(za));
+        return -1;
+    }
+    return 0;
+}
+
+
+static int
+set_file_comment(char *argv[]) {
     zip_uint64_t idx;
     idx = strtoull(argv[0], NULL, 10);
     if (zip_file_set_comment(za, idx, argv[1], (zip_uint16_t)strlen(argv[1]), 0) < 0) {
@@ -472,7 +554,7 @@ set_file_comment(int argc, char *argv[]) {
 }
 
 static int
-set_file_compression(int argc, char *argv[]) {
+set_file_compression(char *argv[]) {
     zip_int32_t method;
     zip_uint32_t flags;
     zip_uint64_t idx;
@@ -487,7 +569,7 @@ set_file_compression(int argc, char *argv[]) {
 }
 
 static int
-set_file_encryption(int argc, char *argv[]) {
+set_file_encryption(char *argv[]) {
     zip_uint16_t method;
     zip_uint64_t idx;
     char *password;
@@ -505,7 +587,7 @@ set_file_encryption(int argc, char *argv[]) {
 }
 
 static int
-set_file_dostime(int argc, char *argv[]) {
+set_file_dostime(char *argv[]) {
     /* set file last modification time (mtime) directly */
     zip_uint16_t dostime, dosdate;
     zip_uint64_t idx;
@@ -520,7 +602,7 @@ set_file_dostime(int argc, char *argv[]) {
 }
 
 static int
-set_file_mtime(int argc, char *argv[]) {
+set_file_mtime(char *argv[]) {
     /* set file last modification time (mtime) */
     time_t mtime;
     zip_uint64_t idx;
@@ -534,7 +616,7 @@ set_file_mtime(int argc, char *argv[]) {
 }
 
 static int
-set_file_mtime_all(int argc, char *argv[]) {
+set_file_mtime_all(char *argv[]) {
     /* set last modification time (mtime) for all files */
     time_t mtime;
     zip_int64_t num_entries;
@@ -555,7 +637,7 @@ set_file_mtime_all(int argc, char *argv[]) {
 }
 
 static int
-set_password(int argc, char *argv[]) {
+set_password(char *argv[]) {
     /* set default password */
     if (zip_set_default_password(za, argv[0]) < 0) {
         fprintf(stderr, "can't set default password to '%s'\n", argv[0]);
@@ -565,7 +647,7 @@ set_password(int argc, char *argv[]) {
 }
 
 static int
-zstat(int argc, char *argv[]) {
+zstat(char *argv[]) {
     zip_uint64_t idx;
     char buf[100];
     struct zip_stat sb;
@@ -577,7 +659,7 @@ zstat(int argc, char *argv[]) {
     }
 
     if (sb.valid & ZIP_STAT_NAME)
-        printf("name: '%s'\n", sb.name);
+        printf("name: '%s'\n", encode_filename(sb.name));
     if (sb.valid & ZIP_STAT_INDEX)
         printf("index: '%" PRIu64 "'\n", sb.index);
     if (sb.valid & ZIP_STAT_SIZE)
@@ -586,12 +668,8 @@ zstat(int argc, char *argv[]) {
         printf("compressed size: '%" PRIu64 "'\n", sb.comp_size);
     if (sb.valid & ZIP_STAT_MTIME) {
         struct tm *tpm;
-#ifdef HAVE_LOCALTIME_R
         struct tm tm;
-        tpm = localtime_r(&sb.mtime, &tm);
-#else
-        tpm = localtime(&sb.mtime);
-#endif
+        tpm = zip_localtime(&sb.mtime, &tm);
         if (tpm == NULL) {
             printf("mtime: <not valid>\n");
         }
@@ -613,6 +691,22 @@ zstat(int argc, char *argv[]) {
     return 0;
 }
 
+static int parse_archive_flag(const char* arg) {
+    if (strcasecmp(arg, "rdonly") == 0) {
+        return ZIP_AFL_RDONLY;
+    }
+    else if (strcasecmp(arg, "is-torrentzip") == 0) {
+        return ZIP_AFL_IS_TORRENTZIP;
+    }
+    else if (strcasecmp(arg, "want-torrentzip") == 0) {
+        return ZIP_AFL_WANT_TORRENTZIP;
+    }
+    else if (strcasecmp(arg, "create-or-keep-file-for-empty-archive") == 0) {
+        return ZIP_AFL_CREATE_OR_KEEP_FILE_FOR_EMPTY_ARCHIVE;
+    }
+    return -1;
+}
+
 static zip_flags_t
 get_flags(const char *arg) {
     zip_flags_t flags = 0;
@@ -626,6 +720,14 @@ get_flags(const char *arg) {
         flags |= ZIP_FL_LOCAL;
     if (strchr(arg, 'u') != NULL)
         flags |= ZIP_FL_UNCHANGED;
+    if (strchr(arg, '8') != NULL)
+        flags |= ZIP_FL_ENC_UTF_8;
+    if (strchr(arg, '4') != NULL)
+        flags |= ZIP_FL_ENC_CP437;
+    if (strchr(arg, 'r') != NULL)
+        flags |= ZIP_FL_ENC_RAW;
+    if (strchr(arg, 's') != NULL)
+        flags |= ZIP_FL_ENC_STRICT;
     return flags;
 }
 
@@ -691,8 +793,6 @@ hexdump(const zip_uint8_t *data, zip_uint16_t len) {
 
     for (i = 0; i < len; i++)
         printf("%02x", data[i]);
-
-    return;
 }
 
 
@@ -733,12 +833,14 @@ dispatch_table_t dispatch_table[] = {{"add", 2, "name content", "add file called
                                      {"add_file", 4, "name file_to_add offset len", "add file to archive, len bytes starting from offset", add_file},
                                      {"add_from_zip", 5, "name archivename index offset len", "add file from another archive, len bytes starting from offset", add_from_zip},
                                      {"cat", 1, "index", "output file contents to stdout", cat},
+                                     {"cat_partial", 3, "index start length", "output partial file contents to stdout", cat_partial},
                                      {"count_extra", 2, "index flags", "show number of extra fields for archive entry", count_extra},
                                      {"count_extra_by_id", 3, "index extra_id flags", "show number of extra fields of type extra_id for archive entry", count_extra_by_id},
                                      {"delete", 1, "index", "remove entry", delete},
                                      {"delete_extra", 3, "index extra_idx flags", "remove extra field", delete_extra},
                                      {"delete_extra_by_id", 4, "index extra_id extra_index flags", "remove extra field of type extra_id", delete_extra_by_id},
                                      {"get_archive_comment", 0, "", "show archive comment", get_archive_comment},
+                                     {"get_archive_flag", 1, "flag", "show archive flag", get_archive_flag},
                                      {"get_extra", 3, "index extra_index flags", "show extra field", get_extra},
                                      {"get_extra_by_id", 4, "index extra_id extra_index flags", "show extra field of type extra_id", get_extra_by_id},
                                      {"get_file_comment", 1, "index", "get file comment", get_file_comment},
@@ -748,6 +850,7 @@ dispatch_table_t dispatch_table[] = {{"add", 2, "name content", "add file called
                                      {"rename", 2, "index name", "rename entry", zrename},
                                      {"replace_file_contents", 2, "index data", "replace entry with data", replace_file_contents},
                                      {"set_archive_comment", 1, "comment", "set archive comment", set_archive_comment},
+                                     {"set_archive_flag", 2, "flag", "set archive flag", set_archive_flag},
                                      {"set_extra", 5, "index extra_id extra_index flags value", "set extra field", set_extra},
                                      {"set_file_comment", 2, "index comment", "set file comment", set_file_comment},
                                      {"set_file_compression", 3, "index method compression_flags", "set file compression method", set_file_compression},
@@ -775,7 +878,7 @@ dispatch(int argc, char *argv[]) {
                 fprintf(stderr, "not enough arguments for command '%s': %d available, %d needed\n", dispatch_table[i].cmdline_name, argc, dispatch_table[i].argument_count);
                 return -1;
             }
-            if (dispatch_table[i].function(argc, argv) == 0)
+            if (dispatch_table[i].function(argv) == 0)
                 return 1 + dispatch_table[i].argument_count;
             return -1;
         }
@@ -826,11 +929,20 @@ usage(const char *progname, const char *reason) {
     }
     fprintf(out, "\nSupported flags are:\n"
                  "\t0\t(no flags)\n"
+                 "\t4\tZIP_FL_ENC_CP437\n"
+                 "\t8\tZIP_FL_ENC_UTF_8\n"
                  "\tC\tZIP_FL_NOCASE\n"
                  "\tc\tZIP_FL_CENTRAL\n"
                  "\td\tZIP_FL_NODIR\n"
                  "\tl\tZIP_FL_LOCAL\n"
+                 "\tr\tZIP_FL_ENC_RAW\n"
+                 "\ts\tZIP_FL_ENC_STRICT\n"
                  "\tu\tZIP_FL_UNCHANGED\n");
+    fprintf(out, "\nSupported archive flags are:\n"
+	         "\tcreate-or-keep-empty-file-for-archive\n"
+	         "\tis-torrentzip\n"
+	         "\trdonly\n"
+	         "\twant-torrentzip\n");
     fprintf(out, "\nSupported compression methods are:\n"
                  "\tdefault\n");
     if (zip_compression_method_supported(ZIP_CM_BZIP2, 1)) {
@@ -956,6 +1068,10 @@ main(int argc, char *argv[]) {
         }
     }
 
+#ifdef PRECLOSE_REGRESS
+    PRECLOSE_REGRESS;
+#endif
+
     if (zip_close(za) == -1) {
         fprintf(stderr, "can't close zip archive '%s': %s\n", archive, zip_strerror(za));
         return 1;
@@ -972,3 +1088,52 @@ main(int argc, char *argv[]) {
 
     return err;
 }
+
+#define BIN2HEX(n) ((n) >= 10 ? (n) + 'a' - 10 : (n) + '0')
+#define HEX2BIN(c) (((c) >= '0' && (c) <= '9') ? (c) - '0' : ((c) >= 'A' && (c) <= 'F') ? (c) - 'A' + 10 : (c) - 'a' + 10)
+
+#define FILENAME_BUFFER_LENGTH 8192
+static char filename_buffer[FILENAME_BUFFER_LENGTH + 1];
+
+static const char* encode_filename(const char* name) {
+    if (!hex_encoded_filenames) {
+        return name;
+    }
+
+    if (strlen(name) > FILENAME_BUFFER_LENGTH / 2) {
+        // TODO: error message
+        exit(1);
+    }
+    const unsigned char* s = (const unsigned char*)name;
+    char* t = filename_buffer;
+    while (*s != '\0') {
+        *(t++) = BIN2HEX(*s >> 4);
+        *(t++) = BIN2HEX(*s & 0xf);
+        s += 1;
+    }
+    *t = '\0';
+    return filename_buffer;
+}
+
+static const char* decode_filename(const char* name) {
+    if (!hex_encoded_filenames) {
+        return name;
+    }
+
+    if (strlen(name) > FILENAME_BUFFER_LENGTH * 2) {
+        // TODO: error message
+        exit(1);
+    }
+    // TODO: check that strlen(name) % 2 == 0
+    // TODO: check with strspn that s is all hex digits
+
+    unsigned char *t = (unsigned char*)filename_buffer;
+    const char *s = name;
+    while (*s != '\0') {
+        *(t++) = (HEX2BIN(s[0]) << 4) | HEX2BIN(s[1]);
+        s += 2;
+    }
+    *t = '\0';
+
+    return filename_buffer;
+}
\ No newline at end of file
